# -*- coding: utf-8 -*-
__title__ = 'pyasista'
__description__ = 'Arcelik Asista Smart Assistant AWS Lambda Python Library'
__url__ = 'https://github.com/Arcelik/pyasista'
__version__ = '0.5.0'
__author__ = 'Arcelik Asista Team'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018 Arcelik A.S.'
